"""
configs of models of CRNN structures, for classification
"""

from copy import deepcopy

from ..cfg import CFG
from ..utils.utils_nn import adjust_cnn_filter_lengths
from .attn import (
    global_context,
    non_local,
    squeeze_excitation,
    transformer,
    self_attention,
)
from .cnn import (
    densenet_leadwise,
    densenet_vanilla,
    multi_scopic,
    multi_scopic_block,
    multi_scopic_leadwise,
    mobilenet_v1_vanilla,
    mobilenet_v2_vanilla,
    mobilenet_v3_small,
    resnet_block_stanford,
    resnet_cpsc2018_leadwise,
    resnet_nature_comm,
    resnet_nature_comm_bottle_neck,
    resnet_nature_comm_bottle_neck_gc,
    resnet_nature_comm_bottle_neck_nl,
    resnet_nature_comm_bottle_neck_se,
    resnet_nature_comm_gc,
    resnet_nature_comm_nl,
    resnet_nature_comm_se,
    resnet_stanford,
    resnet_vanilla_18,
    resnet_vanilla_34,
    resnet_vanilla_50,
    resnetN,
    resnetNB,
    resnetNBS,
    resnetNS,
    resnext_vanilla_50_32x4d,
    tresnetF,
    tresnetM,
    tresnetN,
    tresnetP,
    tresnetS,
    regnet_16_8,
    regnet_27_24,
    regnet_23_168,
    regnet_S,
    vgg16,
    vgg16_leadwise,
    vgg_block_basic,
    vgg_block_mish,
    vgg_block_swish,
    xception_leadwise,
    xception_vanilla,
)
from .mlp import linear
from .rnn import lstm


__all__ = [
    "ECG_CRNN_CONFIG",
]


ECG_CRNN_CONFIG = CFG()
ECG_CRNN_CONFIG.fs = 500

# cnn part
ECG_CRNN_CONFIG.cnn = CFG()
# ECG_CRNN_CONFIG.cnn.name = "resnet_leadwise"
ECG_CRNN_CONFIG.cnn.name = "resnet_nature_comm_bottle_neck"


ECG_CRNN_CONFIG.cnn.vgg16 = deepcopy(vgg16)
ECG_CRNN_CONFIG.cnn.vgg16.block = deepcopy(vgg_block_basic)
ECG_CRNN_CONFIG.cnn.vgg16 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.vgg16, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.vgg16_mish = deepcopy(vgg16)
ECG_CRNN_CONFIG.cnn.vgg16_mish.block = deepcopy(vgg_block_mish)
ECG_CRNN_CONFIG.cnn.vgg16_mish = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.vgg16_mish, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.vgg16_swish = deepcopy(vgg16)
ECG_CRNN_CONFIG.cnn.vgg16_swish.block = deepcopy(vgg_block_swish)
ECG_CRNN_CONFIG.cnn.vgg16_swish = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.vgg16_swish, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.vgg16_leadwise = deepcopy(vgg16_leadwise)
ECG_CRNN_CONFIG.cnn.vgg16_leadwise.block = deepcopy(vgg_block_swish)
ECG_CRNN_CONFIG.cnn.vgg16_leadwise = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.vgg16_leadwise, ECG_CRNN_CONFIG.fs
)
# ECG_CRNN_CONFIG.cnn.vgg16_dilation = deepcopy(vgg16)
# ECG_CRNN_CONFIG.cnn.vgg16_dilation.block = deepcopy(vgg_block_basic)

# ECG_CRNN_CONFIG.cnn.resnet_cpsc2018 = deepcopy(resnet_cpsc2018)
# ECG_CRNN_CONFIG.cnn.resnet_cpsc2018 = adjust_cnn_filter_lengths(ECG_CRNN_CONFIG.cnn.resnet, ECG_CRNN_CONFIG.fs)
ECG_CRNN_CONFIG.cnn.resnet_18 = deepcopy(resnet_vanilla_18)
ECG_CRNN_CONFIG.cnn.resnet_18 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_18, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_34 = deepcopy(resnet_vanilla_34)
ECG_CRNN_CONFIG.cnn.resnet_34 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_34, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_50 = deepcopy(resnet_vanilla_50)
ECG_CRNN_CONFIG.cnn.resnet_50 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_50, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnext_50_32x4d = deepcopy(resnext_vanilla_50_32x4d)
ECG_CRNN_CONFIG.cnn.resnext_50_32x4d = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnext_50_32x4d, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.resnet_nature_comm = deepcopy(resnet_nature_comm)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_se = deepcopy(resnet_nature_comm_se)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_se = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm_se, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_gc = deepcopy(resnet_nature_comm_gc)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_gc = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm_gc, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_nl = deepcopy(resnet_nature_comm_nl)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_nl = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm_nl, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck = deepcopy(
    resnet_nature_comm_bottle_neck
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_se = deepcopy(
    resnet_nature_comm_bottle_neck_se
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_se = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_se, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_gc = deepcopy(
    resnet_nature_comm_bottle_neck_gc
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_gc = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_gc, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_nl = deepcopy(
    resnet_nature_comm_bottle_neck_nl
)
ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_nl = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_nature_comm_bottle_neck_nl, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.resnetN = deepcopy(resnetN)
ECG_CRNN_CONFIG.cnn.resnetN = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnetN, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnetNB = deepcopy(resnetNB)
ECG_CRNN_CONFIG.cnn.resnetNB = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnetNB, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnetNS = deepcopy(resnetNS)
ECG_CRNN_CONFIG.cnn.resnetNS = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnetNS, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.resnetNBS = deepcopy(resnetNBS)
ECG_CRNN_CONFIG.cnn.resnetNBS = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnetNBS, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.tresnetF = deepcopy(tresnetF)
ECG_CRNN_CONFIG.cnn.tresnetF = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.tresnetF, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.tresnetP = deepcopy(tresnetP)
ECG_CRNN_CONFIG.cnn.tresnetP = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.tresnetP, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.tresnetN = deepcopy(tresnetN)
ECG_CRNN_CONFIG.cnn.tresnetN = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.tresnetN, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.tresnetS = deepcopy(tresnetS)
ECG_CRNN_CONFIG.cnn.tresnetS = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.tresnetS, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.tresnetM = deepcopy(tresnetM)
ECG_CRNN_CONFIG.cnn.tresnetM = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.tresnetM, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.resnet_leadwise = deepcopy(resnet_cpsc2018_leadwise)
ECG_CRNN_CONFIG.cnn.resnet_leadwise = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_leadwise, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.resnet_stanford = deepcopy(resnet_stanford)
ECG_CRNN_CONFIG.cnn.resnet_stanford.block = deepcopy(resnet_block_stanford)
ECG_CRNN_CONFIG.cnn.resnet_stanford = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.resnet_stanford, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.regnet_16_8 = deepcopy(regnet_16_8)
ECG_CRNN_CONFIG.cnn.regnet_16_8 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.regnet_16_8, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.regnet_27_24 = deepcopy(regnet_27_24)
ECG_CRNN_CONFIG.cnn.regnet_27_24 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.regnet_27_24, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.regnet_23_168 = deepcopy(regnet_23_168)
ECG_CRNN_CONFIG.cnn.regnet_23_168 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.regnet_23_168, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.regnet_S = deepcopy(regnet_S)
ECG_CRNN_CONFIG.cnn.regnet_S = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.regnet_S, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.multi_scopic = deepcopy(multi_scopic)
ECG_CRNN_CONFIG.cnn.multi_scopic.block = deepcopy(multi_scopic_block)
ECG_CRNN_CONFIG.cnn.multi_scopic = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.multi_scopic, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.multi_scopic_leadwise = deepcopy(multi_scopic_leadwise)
ECG_CRNN_CONFIG.cnn.multi_scopic_leadwise.block = deepcopy(multi_scopic_block)
ECG_CRNN_CONFIG.cnn.multi_scopic_leadwise = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.multi_scopic_leadwise, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.mobilenet_v1 = deepcopy(mobilenet_v1_vanilla)
ECG_CRNN_CONFIG.cnn.mobilenet_v1 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.mobilenet_v1, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.mobilenet_v2 = deepcopy(mobilenet_v2_vanilla)
ECG_CRNN_CONFIG.cnn.mobilenet_v2 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.mobilenet_v2, ECG_CRNN_CONFIG.fs
)
ECG_CRNN_CONFIG.cnn.mobilenet_v3 = deepcopy(mobilenet_v3_small)
ECG_CRNN_CONFIG.cnn.mobilenet_v3 = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.mobilenet_v3, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.xception_vanilla = deepcopy(xception_vanilla)
ECG_CRNN_CONFIG.cnn.xception_leadwise = deepcopy(xception_leadwise)
ECG_CRNN_CONFIG.cnn.xception_leadwise = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.xception_leadwise, ECG_CRNN_CONFIG.fs
)

ECG_CRNN_CONFIG.cnn.densenet_vanilla = deepcopy(densenet_vanilla)
ECG_CRNN_CONFIG.cnn.densenet_leadwise = deepcopy(densenet_leadwise)
ECG_CRNN_CONFIG.cnn.densenet_leadwise = adjust_cnn_filter_lengths(
    ECG_CRNN_CONFIG.cnn.densenet_leadwise, ECG_CRNN_CONFIG.fs
)


# rnn part
ECG_CRNN_CONFIG.rnn = CFG()
ECG_CRNN_CONFIG.rnn.name = "none"  # "none", "lstm", "linear"

ECG_CRNN_CONFIG.rnn.lstm = deepcopy(lstm)
ECG_CRNN_CONFIG.rnn.linear = deepcopy(linear)


# attention part
ECG_CRNN_CONFIG.attn = CFG()
ECG_CRNN_CONFIG.attn.name = "se"  # "none", "se", "gc", "nl"

ECG_CRNN_CONFIG.attn.se = deepcopy(squeeze_excitation)

ECG_CRNN_CONFIG.attn.gc = deepcopy(global_context)

ECG_CRNN_CONFIG.attn.nl = deepcopy(non_local)

ECG_CRNN_CONFIG.attn.transformer = deepcopy(transformer)

ECG_CRNN_CONFIG.attn.sa = deepcopy(self_attention)


# global pooling
# currently is fixed using `AdaptiveMaxPool1d`
ECG_CRNN_CONFIG.global_pool = "max"  # "avg", "attn"
ECG_CRNN_CONFIG.global_pool_size = 1


ECG_CRNN_CONFIG.clf = CFG()
ECG_CRNN_CONFIG.clf.out_channels = [
    1024,
    # not including the last linear layer, whose out channels equals n_classes
]
ECG_CRNN_CONFIG.clf.activation = "mish"
ECG_CRNN_CONFIG.clf.bias = True
ECG_CRNN_CONFIG.clf.kernel_initializer = "he_normal"
ECG_CRNN_CONFIG.clf.dropouts = 0.2



"""
torch_ecg.models
================

This module contains the model architectures for ECG various tasks.

.. contents:: torch_ecg.models
    :depth: 2
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg.models

Convolutional neural backbones
------------------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    ResNet
    RegNet
    VGG16
    Xception
    DenseNet
    MobileNetV1
    MobileNetV2
    MobileNetV3
    MultiScopicCNN

Downstream task models
----------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    ECG_UNET
    ECG_SUBTRACT_UNET
    ECG_CRNN
    ECG_SEQ_LAB_NET
    RR_LSTM

Saliency analysis
-----------------
.. autosummary::
    :toctree: generated/
    :recursive:

    GradCam

"""

from .cnn import (
    ResNet,
    RegNet,
    VGG16,
    Xception,
    DenseNet,
    MobileNetV1,
    MobileNetV2,
    MobileNetV3,
    MultiScopicCNN,
)
from .unets import ECG_UNET, ECG_SUBTRACT_UNET
from .ecg_crnn import ECG_CRNN
from .ecg_seq_lab_net import ECG_SEQ_LAB_NET
from .grad_cam import GradCam
from .rr_lstm import RR_LSTM
from .transformers import Transformer
from . import loss
from . import _nets


__all__ = [
    # CNN backbone
    "ResNet",
    "RegNet",
    "VGG16",
    "Xception",
    "DenseNet",
    "MobileNetV1",
    "MobileNetV2",
    "MobileNetV3",
    "MultiScopicCNN",
    # downstream task models
    "ECG_UNET",
    "ECG_SUBTRACT_UNET",
    "ECG_CRNN",
    "ECG_SEQ_LAB_NET",
    "GradCam",
    "RR_LSTM",
    # etc
    "Transformer",
    "loss",
    "_nets",
]



"""
the model of (attention-based) time-incremental CNN

the cnn layers of this model has a constant kernel size 3,
but keep increasing the number of channels
"""

from copy import deepcopy

from ..cfg import CFG
from ..utils.utils_nn import adjust_cnn_filter_lengths
from .cnn import (
    vgg16,
    vgg16_leadwise,
    vgg_block_basic,
    vgg_block_mish,
    resnet_nature_comm_bottle_neck_se,
)


__all__ = [
    "ATI_CNN_CONFIG",
]


ATI_CNN_CONFIG = CFG()
ATI_CNN_CONFIG.fs = 500


# cnn part
ATI_CNN_CONFIG.cnn = CFG()
ATI_CNN_CONFIG.cnn.name = "vgg16"

ATI_CNN_CONFIG.cnn.vgg16 = adjust_cnn_filter_lengths(vgg16, ATI_CNN_CONFIG.fs)
ATI_CNN_CONFIG.cnn.vgg16.block = deepcopy(vgg_block_basic)

ATI_CNN_CONFIG.cnn.vgg16_mish = adjust_cnn_filter_lengths(vgg16, ATI_CNN_CONFIG.fs)
ATI_CNN_CONFIG.cnn.vgg16_mish.block = deepcopy(vgg_block_mish)

ATI_CNN_CONFIG.cnn.vgg16_leadwise = adjust_cnn_filter_lengths(
    vgg16_leadwise, ATI_CNN_CONFIG.fs
)
ATI_CNN_CONFIG.cnn.vgg16_leadwise.block = deepcopy(vgg_block_mish)

ATI_CNN_CONFIG.cnn.resnet = adjust_cnn_filter_lengths(
    resnet_nature_comm_bottle_neck_se, ATI_CNN_CONFIG.fs
)


# rnn part
ATI_CNN_CONFIG.rnn = CFG()
ATI_CNN_CONFIG.rnn.name = "lstm"

ATI_CNN_CONFIG.rnn.lstm = CFG()
ATI_CNN_CONFIG.rnn.lstm.bias = True
ATI_CNN_CONFIG.rnn.lstm.dropout = 0.2
ATI_CNN_CONFIG.rnn.lstm.bidirectional = True
ATI_CNN_CONFIG.rnn.lstm.retseq = False
ATI_CNN_CONFIG.rnn.lstm.hidden_sizes = [128, 32]

ATI_CNN_CONFIG.attn = CFG()
ATI_CNN_CONFIG.attn.name = "none"

ATI_CNN_CONFIG.global_pool = "none"



"""
"""

from copy import deepcopy
from itertools import repeat

import numpy as np

from ...cfg import CFG


__all__ = [
    # vanilla mobilenets
    "mobilenet_v1_vanilla",
    "mobilenet_v2_vanilla",
    "mobilenet_v3_small",
]


mobilenet_v1_vanilla = CFG()
mobilenet_v1_vanilla.fs = 500
mobilenet_v1_vanilla.groups = 1
mobilenet_v1_vanilla.batch_norm = True
mobilenet_v1_vanilla.activation = "relu6"
mobilenet_v1_vanilla.depth_multiplier = (
    1  # multiplier of number of depthwise convolution output channels
)
mobilenet_v1_vanilla.width_multiplier = (
    1.0  # controls the width (number of filters) of the network
)
mobilenet_v1_vanilla.bias = True
mobilenet_v1_vanilla.ordering = "cba"

_base_num_filters = 12 * 3
mobilenet_v1_vanilla.init_num_filters = _base_num_filters
mobilenet_v1_vanilla.init_filter_lengths = 27
mobilenet_v1_vanilla.init_subsample_lengths = 2

mobilenet_v1_vanilla.entry_flow = CFG()
mobilenet_v1_vanilla.entry_flow.out_channels = [
    # 64, 128, 128, 256, 256
    _base_num_filters * 2,
    _base_num_filters * 4,
    _base_num_filters * 4,
    _base_num_filters * 8,
    _base_num_filters * 8,
    _base_num_filters * 16,
]
mobilenet_v1_vanilla.entry_flow.filter_lengths = 15
mobilenet_v1_vanilla.entry_flow.subsample_lengths = [
    1,
    2,
    1,
    2,
    1,
    2,
]
mobilenet_v1_vanilla.entry_flow.groups = mobilenet_v1_vanilla.groups
mobilenet_v1_vanilla.entry_flow.batch_norm = mobilenet_v1_vanilla.batch_norm
mobilenet_v1_vanilla.entry_flow.activation = mobilenet_v1_vanilla.activation

mobilenet_v1_vanilla.middle_flow = CFG()
mobilenet_v1_vanilla.middle_flow.out_channels = list(repeat(_base_num_filters * 16, 5))
mobilenet_v1_vanilla.middle_flow.filter_lengths = 13
mobilenet_v1_vanilla.middle_flow.subsample_lengths = 1
mobilenet_v1_vanilla.middle_flow.groups = mobilenet_v1_vanilla.groups
mobilenet_v1_vanilla.middle_flow.batch_norm = mobilenet_v1_vanilla.batch_norm
mobilenet_v1_vanilla.middle_flow.activation = mobilenet_v1_vanilla.activation

mobilenet_v1_vanilla.exit_flow = CFG()
mobilenet_v1_vanilla.exit_flow.out_channels = [
    _base_num_filters * 32,
    _base_num_filters * 32,
]
mobilenet_v1_vanilla.exit_flow.filter_lengths = 17
mobilenet_v1_vanilla.exit_flow.subsample_lengths = [2, 1]
mobilenet_v1_vanilla.exit_flow.groups = mobilenet_v1_vanilla.groups
mobilenet_v1_vanilla.exit_flow.batch_norm = mobilenet_v1_vanilla.batch_norm
mobilenet_v1_vanilla.exit_flow.activation = mobilenet_v1_vanilla.activation


mobilenet_v2_vanilla = CFG()
mobilenet_v2_vanilla.fs = 500
mobilenet_v2_vanilla.groups = 1
mobilenet_v2_vanilla.batch_norm = True
mobilenet_v2_vanilla.activation = "relu6"
mobilenet_v2_vanilla.depth_multiplier = (
    1  # multiplier of number of depthwise convolution output channels
)
mobilenet_v2_vanilla.width_multiplier = (
    1.0  # controls the width (number of filters) of the network
)
mobilenet_v2_vanilla.bias = True
mobilenet_v2_vanilla.ordering = "cba"

_base_num_filters = 12
mobilenet_v2_vanilla.stem = CFG(
    num_filters=_base_num_filters * 4,
    filter_lengths=27,
    subsample_lengths=2,
)

_inverted_residual_setting = np.array(
    [
        # t, c, n, s, k
        [1, _base_num_filters * 2, 1, 1, 15],
        [6, _base_num_filters * 3, 2, 2, 15],
        [6, _base_num_filters * 4, 3, 2, 15],
        [6, _base_num_filters * 6, 4, 2, 15],
        [6, _base_num_filters * 8, 3, 1, 15],
        [6, _base_num_filters * 20, 3, 2, 15],
        [6, _base_num_filters * 40, 1, 1, 15],
        # t: expansion
        # c: output channels
        # n: number of blocks
        # s: stride
        # k: kernel size
    ]
).astype(int)
mobilenet_v2_vanilla.inv_res = CFG()
mobilenet_v2_vanilla.inv_res.expansions = _inverted_residual_setting[..., 0].tolist()
mobilenet_v2_vanilla.inv_res.out_channels = _inverted_residual_setting[..., 1].tolist()
mobilenet_v2_vanilla.inv_res.n_blocks = _inverted_residual_setting[..., 2].tolist()
mobilenet_v2_vanilla.inv_res.strides = _inverted_residual_setting[..., 3].tolist()
mobilenet_v2_vanilla.inv_res.filter_lengths = _inverted_residual_setting[
    ..., 4
].tolist()

mobilenet_v2_vanilla.exit_flow = CFG(
    num_filters=_base_num_filters * 128,
    filter_lengths=19,
    subsample_lengths=2,
)


# mobilenet_v3_small = deepcopy(mobilenet_v2_vanilla)
mobilenet_v3_small = CFG()
mobilenet_v3_small.fs = 500
mobilenet_v3_small.groups = 1
mobilenet_v3_small.batch_norm = True
mobilenet_v3_small.activation = "relu"
mobilenet_v3_small.depth_multiplier = (
    1  # multiplier of number of depthwise convolution output channels
)
mobilenet_v3_small.width_multiplier = (
    1.0  # controls the width (number of filters) of the network
)
mobilenet_v3_small.bias = True
mobilenet_v3_small.ordering = "cba"

_base_num_filters = 12
mobilenet_v3_small.stem = CFG(
    # num_filters=[_base_num_filters * 2] * 2,
    num_filters=[_base_num_filters * 2],
    filter_lengths=27,
    subsample_lengths=2,
)

_inverted_residual_setting = np.array(
    [
        # n, t, c, s, k, d
        [1, 4, _base_num_filters * 4, 2, 11, 1],
        [2, 4, _base_num_filters * 8, 2, 11, 1],
        [5, 4, _base_num_filters * 16, 2, 17, 1],
        [3, 4, _base_num_filters * 32, 2, 17, 1],
        # n: number of blocks
        # t: expansion
        # c: output channels
        # s: stride
        # k: kernel size
        # d: dilation
    ]
).astype(int)
_attn_setting = CFG(
    name="se",
    pos=2,
    reduction=8,  # not including the last linear layer
    activation="relu",
    kw_activation=CFG(inplace=True),
)
mobilenet_v3_small.inv_res = CFG()
mobilenet_v3_small.inv_res.n_blocks = _inverted_residual_setting[..., 0].tolist()
mobilenet_v3_small.inv_res.expansions = _inverted_residual_setting[..., 1].tolist()
mobilenet_v3_small.inv_res.out_channels = _inverted_residual_setting[..., 2].tolist()
mobilenet_v3_small.inv_res.strides = _inverted_residual_setting[..., 3].tolist()
mobilenet_v3_small.inv_res.filter_lengths = _inverted_residual_setting[..., 4].tolist()
mobilenet_v3_small.inv_res.dilations = _inverted_residual_setting[..., 5].tolist()
mobilenet_v3_small.inv_res.activations = [
    "relu",
    "relu",
    "hardswish",
    "hardswish",
]
mobilenet_v3_small.inv_res.groups = 1
mobilenet_v3_small.inv_res.attns = [
    deepcopy(_attn_setting),
    None,
    deepcopy(_attn_setting),
    deepcopy(_attn_setting),
]

mobilenet_v3_small.exit_flow = CFG(
    num_filters=_base_num_filters * 128,
    filter_lengths=19,
    subsample_lengths=2,
)



"""
torch_ecg.utils
===============

This module contains a collection of utility functions and classes that are used
throughout the package.

.. contents:: torch_ecg.utils
    :depth: 2
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg.utils

Neural network auxiliary functions and classes
----------------------------------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    extend_predictions
    compute_output_shape
    compute_conv_output_shape
    compute_deconv_output_shape
    compute_maxpool_output_shape
    compute_avgpool_output_shape
    compute_sequential_output_shape
    compute_module_size
    default_collate_fn
    compute_receptive_field
    adjust_cnn_filter_lengths
    SizeMixin
    CkptMixin

Signal processing functions
---------------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    smooth
    resample_irregular_timeseries
    detect_peaks
    remove_spikes_naive
    butter_bandpass_filter
    get_ampl
    normalize
    normalize_t
    resample_t

Data operations
---------------
.. autosummary::
    :toctree: generated/
    :recursive:

    get_mask
    class_weight_to_sample_weight
    ensure_lead_fmt
    ensure_siglen
    masks_to_waveforms
    mask_to_intervals
    uniform
    stratified_train_test_split
    cls_to_bin
    generate_weight_mask

Interval operations
-------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    overlaps
    validate_interval
    in_interval
    in_generalized_interval
    intervals_union
    generalized_intervals_union
    intervals_intersection
    generalized_intervals_intersection
    generalized_interval_complement
    get_optimal_covering
    interval_len
    generalized_interval_len
    find_extrema
    is_intersect
    max_disjoint_covering

Metrics computations
--------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    top_n_accuracy
    confusion_matrix
    ovr_confusion_matrix
    metrics_from_confusion_matrix
    compute_wave_delineation_metrics
    QRS_score

Decorators and Mixins
---------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    add_docstring
    remove_parameters_returns_from_docstring
    default_class_repr
    ReprMixin
    CitationMixin
    get_kwargs
    get_required_args
    add_kwargs

Path operations
---------------
.. autosummary::
    :toctree: generated/
    :recursive:

    get_record_list_recursive3

String operations
-----------------
.. autosummary::
    :toctree: generated/
    :recursive:

    dict_to_str
    str2bool
    nildent
    get_date_str

Miscellaneous
-------------
.. autosummary::
    :toctree: generated/
    :recursive:

    init_logger
    list_sum
    dicts_equal
    MovingAverage
    Timer
    timeout

"""
from . import ecg_arrhythmia_knowledge as EAK
from .download import http_get
from .misc import (
    get_record_list_recursive3,
    dict_to_str,
    str2bool,
    init_logger,
    get_date_str,
    list_sum,
    dicts_equal,
    default_class_repr,
    ReprMixin,
    CitationMixin,
    MovingAverage,
    nildent,
    add_docstring,
    remove_parameters_returns_from_docstring,
    timeout,
    Timer,
    get_kwargs,
    get_required_args,
    add_kwargs,
)
from .utils_data import (
    get_mask,
    class_weight_to_sample_weight,
    ensure_lead_fmt,
    ensure_siglen,
    ECGWaveForm,
    ECGWaveFormNames,
    masks_to_waveforms,
    mask_to_intervals,
    uniform,
    stratified_train_test_split,
    cls_to_bin,
    generate_weight_mask,
)
from .utils_interval import (
    overlaps,
    validate_interval,
    in_interval,
    in_generalized_interval,
    intervals_union,
    generalized_intervals_union,
    intervals_intersection,
    generalized_intervals_intersection,
    generalized_interval_complement,
    get_optimal_covering,
    interval_len,
    generalized_interval_len,
    find_extrema,
    is_intersect,
    max_disjoint_covering,
)
from .utils_metrics import (
    top_n_accuracy,
    confusion_matrix,
    ovr_confusion_matrix,
    QRS_score,
    metrics_from_confusion_matrix,
    compute_wave_delineation_metrics,
)
from .utils_nn import (
    extend_predictions,
    compute_output_shape,
    compute_conv_output_shape,
    compute_deconv_output_shape,
    compute_maxpool_output_shape,
    compute_avgpool_output_shape,
    compute_sequential_output_shape,
    compute_module_size,
    default_collate_fn,
    compute_receptive_field,
    adjust_cnn_filter_lengths,
    SizeMixin,
    CkptMixin,
)
from .utils_signal import (
    smooth,
    resample_irregular_timeseries,
    detect_peaks,
    remove_spikes_naive,
    butter_bandpass_filter,
    get_ampl,
    normalize,
)
from .utils_signal_t import (
    normalize as normalize_t,
    resample as resample_t,
)


__all__ = [
    "EAK",
    "http_get",
    "get_record_list_recursive3",
    "dict_to_str",
    "str2bool",
    "init_logger",
    "get_date_str",
    "list_sum",
    "dicts_equal",
    "default_class_repr",
    "ReprMixin",
    "CitationMixin",
    "MovingAverage",
    "nildent",
    "add_docstring",
    "remove_parameters_returns_from_docstring",
    "timeout",
    "Timer",
    "get_kwargs",
    "get_required_args",
    "add_kwargs",
    "get_mask",
    "class_weight_to_sample_weight",
    "ensure_lead_fmt",
    "ensure_siglen",
    "ECGWaveForm",
    "ECGWaveFormNames",
    "masks_to_waveforms",
    "mask_to_intervals",
    "uniform",
    "stratified_train_test_split",
    "cls_to_bin",
    "generate_weight_mask",
    "overlaps",
    "validate_interval",
    "in_interval",
    "in_generalized_interval",
    "intervals_union",
    "generalized_intervals_union",
    "intervals_intersection",
    "generalized_intervals_intersection",
    "generalized_interval_complement",
    "get_optimal_covering",
    "interval_len",
    "generalized_interval_len",
    "find_extrema",
    "is_intersect",
    "max_disjoint_covering",
    "top_n_accuracy",
    "confusion_matrix",
    "ovr_confusion_matrix",
    "QRS_score",
    "metrics_from_confusion_matrix",
    "compute_wave_delineation_metrics",
    "extend_predictions",
    "compute_output_shape",
    "compute_conv_output_shape",
    "compute_deconv_output_shape",
    "compute_maxpool_output_shape",
    "compute_avgpool_output_shape",
    "compute_sequential_output_shape",
    "compute_module_size",
    "default_collate_fn",
    "compute_receptive_field",
    "adjust_cnn_filter_lengths",
    "SizeMixin",
    "CkptMixin",
    "smooth",
    "resample_irregular_timeseries",
    "detect_peaks",
    "remove_spikes_naive",
    "butter_bandpass_filter",
    "get_ampl",
    "normalize",
    "normalize_t",
    "resample_t",
]



# -*- coding: utf-8 -*-

from collections import defaultdict, Counter
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import pandas as pd
import wfdb
from tqdm.auto import tqdm

from ...cfg import CFG, DEFAULTS
from ...utils.misc import get_record_list_recursive3, add_docstring
from ...utils.utils_interval import generalized_intervals_intersection
from ..base import (
    BeatAnn,
    DataBaseInfo,
    PhysioNetDataBase,
    WFDB_Beat_Annotations,
    WFDB_Non_Beat_Annotations,
    WFDB_Rhythm_Annotations,
)


__all__ = [
    "MITDB",
]


_MITDB_INFO = DataBaseInfo(
    title="""
    MIT-BIH Arrhythmia Database
    """,
    about="""
    1. contains 48 half-hour excerpts of two-channel ambulatory ECG recordings, obtained from 47 subjects.
    2. recordings were digitized at 360 samples per second per channel with 11-bit resolution over a 10 mV range.
    3. annotations contains:

        - beat-wise or finer (e.g. annotations of flutter wave) annotations, accessed via the `symbol` attribute of an `Annotation`.
        - rhythm annotations, accessed via the `aux_note` attribute of an `Annotation`.
    4. Webpage of the database on PhysioNet [1]_.
    """,
    usage=[
        "Beat classification",
        "Rhythm classification (segmentation)",
        "R peaks detection",
    ],
    references=[
        "https://physionet.org/content/mitdb/",
    ],
    doi=[
        "10.1109/51.932724",
        "10.13026/C2F305",
    ],
)


@add_docstring(_MITDB_INFO.format_database_docstring(), mode="prepend")
class MITDB(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "MITDB"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="mitdb",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = 360
        self.data_ext = "dat"
        self.data_pattern = "^[\\d]{3}$"
        self.data_pattern_with_ext = f"^[\\d]{{3}}\\.{self.data_ext}$"
        self.ann_ext = "atr"

        self.beat_types_extended = list("""!"+/AEFJLNQRSV[]aefjx|~""")
        self.nonbeat_types = [
            item
            for item in self.beat_types_extended
            if item in WFDB_Non_Beat_Annotations
        ]
        self.beat_types = [
            item for item in self.beat_types_extended if item in WFDB_Beat_Annotations
        ]
        self.beat_types_map = {item: i for i, item in enumerate(self.beat_types)}
        self.beat_types_extended_map = {
            item: i for i, item in enumerate(self.beat_types_extended)
        }
        self.rhythm_types = [
            "(AB",
            "(AFIB",
            "(AFL",
            "(B",
            "(BII",
            "(IVR",
            "(N",
            "(NOD",
            "(P",
            "(PREX",
            "(SBR",
            "(SVTA",
            "(T",
            "(VFL",
            "(VT",
            "MISSB",
            "PSE",
            "TS",
        ]
        self.rhythm_types = [
            rt.lstrip("(") for rt in self.rhythm_types if rt in WFDB_Rhythm_Annotations
        ]
        self.rhythm_types_map = {rt: idx for idx, rt in enumerate(self.rhythm_types)}
        self._rhythm_ignore_index = -100

        # records have different lead names
        # therefore, self.all_leads should not be set
        # otherwise, it will cause problems when loading data using `self.load_data`
        self._all_leads = ["MLII", "V1", "V2", "V4", "V5"]

        self._ls_rec()

        self._stats = pd.DataFrame()
        self._stats_columns = ["record", "beat_num", "beat_type_num", "rhythm_len"]
        self._aggregate_stats()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        subsample = self._subsample
        self._subsample = None  # so that no subsampling in super()._ls_rec()
        super()._ls_rec()
        # filters out records with names not matching `self.data_pattern`
        if len(self._df_records) > 0:
            self._df_records = self._df_records[
                self._df_records.index.str.match(self.data_pattern)
            ]
        if len(self._all_records) == 0:
            self._df_records = pd.DataFrame()
            self._df_records["path"] = get_record_list_recursive3(
                self.db_dir, self.data_pattern_with_ext, relative=False
            )
            self._df_records["record"] = self._df_records["path"].apply(
                lambda x: x.stem
            )
            self._df_records.set_index("record", inplace=True)
        if subsample is not None:
            size = min(
                len(self._df_records),
                max(1, int(round(subsample * len(self._df_records)))),
            )
            self.logger.debug(
                f"subsample `{size}` records from `{len(self._df_records)}`"
            )
            self._df_records = self._df_records.sample(
                n=size, random_state=DEFAULTS.SEED, replace=False
            )
        self._all_records = self._df_records.index.tolist()
        self._subsample = subsample

    def _aggregate_stats(self) -> None:
        """Aggregate statistics for all records in the database."""
        self._stats = pd.DataFrame(columns=self._stats_columns)
        if len(self) == 0:
            return
        with tqdm(
            range(len(self)),
            desc="Aggregating stats",
            unit="record",
            dynamic_ncols=True,
            mininterval=1.0,
            disable=(self.verbose < 1),
        ) as pbar:
            for idx in pbar:
                rec_ann = self.load_ann(idx)
                beat_type_num = {
                    k: v
                    for k, v in Counter(
                        [item.symbol for item in rec_ann["beat"]]
                    ).most_common()
                }
                beat_num = sum(beat_type_num.values())
                rhythm_len = {
                    k: sum([itv[1] - itv[0] for itv in v])
                    for k, v in rec_ann["rhythm"].items()
                }
                self._stats = pd.concat(
                    [
                        self._stats,
                        pd.DataFrame(
                            [
                                [
                                    self._all_records[idx],
                                    beat_num,
                                    beat_type_num,
                                    rhythm_len,
                                ]
                            ],
                            columns=self._stats_columns,
                        ),
                    ],
                    ignore_index=True,
                )

    def load_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rhythm_format: str = "intervals",
        rhythm_types: Optional[Sequence[str]] = None,
        beat_format: str = "beat",
        beat_types: Optional[Sequence[str]] = None,
        keep_original: bool = False,
    ) -> dict:
        """Load rhythm and beat annotations of the record.

        Rhythm and beat annotations are stored in the `aux_note`, `symbol`
        attributes of corresponding annotation files.
        NOTE that qrs annotations (.qrs files) do NOT contain any rhythm annotations.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        rhythm_format : {"interval", "mask"}, optional
            Format of returned annotation, by default "interval",
            case insensitive.
        rhythm_types : list of str, optional
            Defaults to `self.rhythm_types`.
            If is not None, only the rhythm annotations
            with the specified types will be returned.
        beat_format : {"beat", "dict"}, optional
            Format of returned annotation, by default "beat",
            case insensitive.
        beat_types : List[str], optional
            Beat types to be loaded, by default `self.beat_types`.
            If is not None, only the beat annotations
            with the specified types will be returned.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : dict
            The annotations of ``rhythm`` and ``beat``, with
            ``rhythm`` annotatoins in the format of intervals, or mask;
            ``beat`` annotations in the format of dict or
            :class:`~torch_ecg.databases.BeatAnn`.

        """
        assert rhythm_format.lower() in [
            "intervals",
            "mask",
        ], f"`rhythm_format` must be one of ['intervals', 'mask'], got {rhythm_format}"
        assert beat_format.lower() in [
            "beat",
            "dict",
        ], f"`beat_format` must be one of ['beat', 'dict'], got {beat_format}"
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=self.ann_ext)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        sample_inds = wfdb_ann.sample
        indices = np.where((sample_inds >= sf) & (sample_inds < st))[0]

        if beat_types is None:
            beat_types = self.beat_types

        beat_ann = [
            BeatAnn(i, s)
            for i, s in zip(sample_inds[indices], np.array(wfdb_ann.symbol)[indices])
            if s in beat_types
        ]

        if rhythm_types is None:
            rhythm_types = self.rhythm_types
            rhythm_types_map = self.rhythm_types_map
        else:
            rhythm_types = [rt.lstrip("(") for rt in rhythm_types]
            rhythm_types_map = {rt: idx for idx, rt in enumerate(rhythm_types)}

        rhythm_intervals = defaultdict(list)
        start_idx, rhythm = None, None
        for ra, si in zip(wfdb_ann.aux_note, sample_inds):
            ra = ra.rstrip("\x00").lstrip("(")
            if ra in rhythm_types:
                if start_idx is not None:
                    rhythm_intervals[rhythm].append([start_idx, si])
                start_idx = si
                rhythm = ra.lstrip("(")
        if start_idx is not None:
            rhythm_intervals[rhythm].append([start_idx, si])
        rhythm_intervals = {
            k: np.array(generalized_intervals_intersection(v, [[sf, st]]))
            for k, v in rhythm_intervals.items()
        }
        if rhythm_format.lower() == "mask":
            rhythm_mask = np.full((st - sf,), self._rhythm_ignore_index, dtype=int)
            for k, v in rhythm_intervals.items():
                for itv in v:
                    rhythm_mask[itv[0] - sf : itv[1] - sf] = self.rhythm_types_map[k]

        if not keep_original:
            rhythm_intervals = {k: v - sf for k, v in rhythm_intervals.items()}
            for b in beat_ann:
                b.index -= sf

        # if not extended_beats:
        #     beat_ann = [b for b in beat_ann if b.symbol in self.beat_types]

        if beat_format.lower() == "dict":
            beat_ann = {
                s: np.array([b.index for b in beat_ann if b.symbol == s], dtype=int)
                for s in self.beat_types_extended
            }
            beat_ann = {k: v for k, v in beat_ann.items() if len(v) > 0}

        ann = {}
        ann["beat"] = beat_ann
        if rhythm_format.lower() == "intervals":
            ann["rhythm"] = rhythm_intervals
        else:
            ann["rhythm"] = rhythm_mask
        return ann

    def load_rhythm_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rhythm_format: str = "intervals",
        rhythm_types: Optional[Sequence[str]] = None,
        keep_original: bool = False,
    ) -> Union[Dict[str, list], np.ndarray]:
        """Load rhythm annotations of the record.

        Rhythm annotations are stored in the `aux_note` attribute
        of corresponding annotation files.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        rhythm_format : {"interval", "mask"}, optional
            Format of returned annotation, by default "interval",
            case insensitive.
        rhythm_types : list of str, optional
            Defaults to `self.rhythm_types`.
            If is not None, only the rhythm annotations
            with the specified types will be returned.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : dict or numpy.ndarray
            Annotations in the format of intervals or mask.

        """
        return self.load_ann(
            rec,
            sampfrom,
            sampto,
            rhythm_format=rhythm_format,
            rhythm_types=rhythm_types,
            keep_original=keep_original,
        )["rhythm"]

    def load_beat_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        beat_format: str = "beat",
        beat_types: Optional[Sequence[str]] = None,
        keep_original: bool = False,
    ) -> Union[Dict[str, np.ndarray], List[BeatAnn]]:
        """Load beat annotations of the record.

        Beat annotations are stored in the `symbol` attribute
        of corresponding annotation files.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        beat_format : {"beat", "dict"}, optional
            Format of returned annotation, by default "beat",
            case insensitive.
        beat_types : List[str], optional
            Beat types to be loaded, by default `self.beat_types`.
            If is not None, only the beat annotations
            with the specified types will be returned.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : dict or list
            Locations (indices) of the all the
            beat types ("A", "N", "Q", "V",).

        """
        return self.load_ann(
            rec,
            sampfrom,
            sampto,
            beat_format=beat_format,
            beat_types=beat_types,
            keep_original=keep_original,
        )["beat"]

    def load_rpeak_indices(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
    ) -> np.ndarray:
        """Load rpeak indices of the record.

        Rpeak indices, or equivalently qrs complex locations,
        are stored in the `symbol` attribute of corresponding annotation files,
        regardless of their beat types.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        rpeak_inds : numpy.ndarray
            Locations (indices) of the all the rpeaks (qrs complexes).

        """
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=self.ann_ext)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        rpeak_inds = wfdb_ann.sample
        indices = np.where(
            (rpeak_inds >= sf)
            & (rpeak_inds < st)
            & (np.isin(wfdb_ann.symbol, self.beat_types))
        )[0]
        rpeak_inds = rpeak_inds[indices]
        if not keep_original:
            rpeak_inds -= sf
        return rpeak_inds

    def _get_lead_names(self, rec: Union[str, int]) -> List[str]:
        """Get the names of the leads contained in the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        List[str]
            A list of names of the leads contained in the record.

        """
        return wfdb.rdheader(str(self.get_absolute_path(rec))).sig_name

    @property
    def df_stats(self) -> pd.DataFrame:
        """DataFrame of the statistics of the dataset."""
        if self._stats.empty:
            self._aggregate_stats()
        return self._stats

    @property
    def df_stats_expanded(self) -> pd.DataFrame:
        """Expanded DataFrame of the statistics of the dataset."""
        df = self.df_stats.copy(deep=True)
        for bt in self.beat_types:
            df[f"beat_{bt}"] = df["beat_type_num"].apply(lambda d: d.get(bt, 0))
        for rt in self.rhythm_types:
            df[f"rhythm_{rt}"] = df["rhythm_len"].apply(lambda d: d.get(rt, 0))
        return df.drop(columns=["beat_num", "beat_type_num", "rhythm_len"])

    @property
    def df_stats_expanded_boolean(self) -> pd.DataFrame:
        """Expanded DataFrame of the statistics of the dataset,
        with boolean values.
        """
        df = self.df_stats_expanded.copy(deep=True)
        for col in df.columns:
            if col == "record":
                continue
            df[col] = df[col].apply(lambda x: int(x > 0))
        return df

    @property
    def db_stats(self) -> Dict[str, Dict[str, int]]:
        """Dictionary of the statistics of the dataset."""
        if self._stats.empty:
            self._aggregate_stats()
        rhythm_len = defaultdict(int)
        for rl_dict in self._stats["rhythm_len"]:
            for k, v in rl_dict.items():
                rhythm_len[k] += v
        beat_type_num = defaultdict(int)
        for btn_dict in self._stats["beat_type_num"]:
            for k, v in btn_dict.items():
                beat_type_num[k] += v
        return CFG(rhythm_len=dict(rhythm_len), beat_type_num=dict(beat_type_num))

    def _categorize_records(self, by: str) -> Dict[str, List[str]]:
        """Categorize records by specific attributes.

        Parameters
        ----------
        by : {"beat", "rhythm"}
            The attribute to categorize the records,
            case insensitive.

        Returns
        -------
        dict
            A dict of lists of record names.

        """
        assert by.lower() in [
            "beat",
            "rhythm",
        ], f"`by` should be one of 'beat' or 'rhythm', but got {by}"
        key = dict(beat="beat_type_num", rhythm="rhythm_len")[by.lower()]
        return CFG(
            {
                item: [
                    row["record"]
                    for _, row in self.df_stats.iterrows()
                    if item in row[key]
                ]
                for item in self.db_stats[key]
            }
        )

    @property
    def beat_types_records(self) -> Dict[str, List[str]]:
        """Dictionary of records with specific beat types."""
        return self._categorize_records("beat")

    @property
    def rhythm_types_records(self) -> Dict[str, List[str]]:
        """Dictionary of records with specific rhythm types."""
        return self._categorize_records("rhythm")

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ann: Optional[Dict[str, np.ndarray]] = None,
        beat_ann: Optional[Dict[str, np.ndarray]] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[int, List[int]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        same_range: bool = False,
        **kwargs: Any,
    ) -> None:
        """Not implemented."""
        raise NotImplementedError

    @property
    def database_info(self) -> DataBaseInfo:
        return _MITDB_INFO



"""
The most frequently used (can serve as baseline) CNN family of physiological signal processing,
whose performance however seems exceeded by newer networks
"""

import textwrap
from copy import deepcopy
from itertools import repeat
from numbers import Real
from typing import Optional, Sequence, Union, List

import torch.nn.functional as F
from torch import Tensor, nn

from ...cfg import CFG
from ...models._nets import (
    Activations,
    Conv_Bn_Activation,
    DownSample,
    SpaceToDepth,
    ZeroPadding,
    make_attention_layer,
)
from ...utils.misc import add_docstring, CitationMixin
from ...utils.utils_nn import (
    SizeMixin,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)


__all__ = [
    "ResNet",
    "ResNetBasicBlock",
    "ResNetBottleNeck",
]


if not hasattr(nn, "Dropout1d"):
    nn.Dropout1d = nn.Dropout  # added in pytorch 1.12


_DEFAULT_BLOCK_CONFIG = {
    "increase_channels_method": "conv",
    "subsample_mode": "conv",
    "activation": "relu",
    "kw_activation": {"inplace": True},
    "kernel_initializer": "he_normal",
    "kw_initializer": {},
    "bias": False,
}


class ResNetBasicBlock(nn.Module, SizeMixin):
    """Building blocks for :class:`ResNet`.

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input tensor.
    num_filters : int
        Number of filters for the convolutional layers.
    filter_length : int
        Length (size) of the filter kernels.
    subsample_lengths : int
        Subsample length, including pool size for short cut,
        and stride for the top convolutional layer.
    groups : int, default 1
        Pattern of connections between inputs and outputs.
        For more details, ref. :class:`torch.nn.Conv1d`.
    dilation : int, default 1
        Not used.
    attn : dict, optional
        Attention mechanism for the neck conv layer.
        If is None, no attention mechanism is used.
        keys:

            - name: str,
              can be "se", "gc", "nl" (alias "nonlocal", "non-local"), etc.
            - pos: int,
              position of the attention mechanism.

        Other keys are specific to the attention mechanism.
    config : dict
        Other hyper-parameters, including
        increase channel method, subsample method, dropouts,
        activation choices, weight initializer, and short cut patterns, etc.

    """

    __name__ = "ResNetBasicBlock"
    expansion = 1  # not used
    __DEFAULT_BASE_WIDTH__ = 12 * 4  # not used
    __DEFAULT_CONFIG__ = _DEFAULT_BLOCK_CONFIG.copy()

    def __init__(
        self,
        in_channels: int,
        num_filters: int,
        filter_length: int,
        subsample_length: int,
        groups: int = 1,
        dilation: int = 1,
        attn: Optional[dict] = None,
        **config,
    ) -> None:
        super().__init__()
        if dilation > 1:
            raise NotImplementedError(f"Dilation > 1 not supported in {self.__name__}")
        self.__num_convs = 2
        self.__in_channels = in_channels
        self.__out_channels = num_filters
        self.__kernel_size = filter_length
        self.__down_scale = subsample_length
        self.__stride = subsample_length
        self.__groups = groups
        self.config = CFG(self.__DEFAULT_CONFIG__.copy())
        self.config.update(config)

        if (
            self.config.increase_channels_method.lower() == "zero_padding"
            and self.__groups != 1
        ):
            raise ValueError(
                "zero padding for increasing channels can not be used with groups != 1"
            )

        self.__attn = attn
        if self.__attn:
            self.__attn = CFG(self.__attn)

        self.__increase_channels = self.__out_channels > self.__in_channels
        self.shortcut = self._make_shortcut_layer()

        self.main_stream = nn.Sequential()
        conv_in_channels = self.__in_channels
        dropout_config = self.config.get("dropout", 0)
        for i in range(self.__num_convs):
            conv_activation = (
                self.config.activation if i < self.__num_convs - 1 else None
            )
            if self.__attn and self.__attn["pos"] == i:
                self.main_stream.add_module(
                    self.__attn["name"],
                    make_attention_layer(conv_in_channels, **self.__attn),
                )
            self.main_stream.add_module(
                f"cba_{i}",
                Conv_Bn_Activation(
                    in_channels=conv_in_channels,
                    out_channels=self.__out_channels,
                    kernel_size=self.__kernel_size,
                    stride=(self.__stride if i == 0 else 1),
                    groups=self.__groups,
                    norm=True,
                    activation=conv_activation,
                    kw_activation=self.config.kw_activation,
                    kernel_initializer=self.config.kernel_initializer,
                    kw_initializer=self.config.kw_initializer,
                    bias=self.config.bias,
                    conv_type=self.config.get("conv_type", None),
                ),
            )
            if i < self.__num_convs - 1:
                if isinstance(dropout_config, dict):
                    if dropout_config["type"] == "1d" and dropout_config["p"] > 0:
                        self.main_stream.add_module(
                            f"dropout_{i}", nn.Dropout1d(dropout_config["p"])
                        )
                    elif dropout_config["type"] is None and dropout_config["p"] > 0:
                        self.main_stream.add_module(
                            f"dropout_{i}", nn.Dropout(dropout_config["p"])
                        )
                elif dropout_config > 0:  # float
                    self.main_stream.add_module(
                        f"dropout_{i}", nn.Dropout(dropout_config)
                    )
            conv_in_channels = self.__out_channels
        if self.__attn and self.__attn["pos"] == -1:
            self.main_stream.add_module(
                self.__attn["name"],
                make_attention_layer(conv_in_channels, **self.__attn),
            )

        if isinstance(self.config.activation, str):
            self.out_activation = Activations[self.config.activation.lower()](
                **self.config.kw_activation
            )
        else:
            self.out_activation = self.config.activation(**self.config.kw_activation)

        if isinstance(dropout_config, dict):
            if dropout_config["type"] == "1d" and dropout_config["p"] > 0:
                self.out_dropout = nn.Dropout1d(dropout_config["p"])
            elif dropout_config["type"] is None and dropout_config["p"] > 0:
                self.out_dropout = nn.Dropout(dropout_config["p"])
        elif dropout_config > 0:  # float
            self.out_dropout = nn.Dropout(self.config.dropout)
        else:
            self.out_dropout = nn.Identity()

    def _make_shortcut_layer(self) -> Union[nn.Module, None]:
        """Make shortcut layer for residual connection."""
        if self.__down_scale > 1 or self.__increase_channels:
            if self.config.increase_channels_method.lower() == "conv":
                shortcut = DownSample(
                    down_scale=self.__down_scale,
                    in_channels=self.__in_channels,
                    out_channels=self.__out_channels,
                    groups=self.__groups,
                    norm=True,
                    mode=self.config.subsample_mode,
                    filt_size=self.config.get("filt_size", 3),  # for blur pool
                )
            elif self.config.increase_channels_method.lower() == "zero_padding":
                batch_norm = (
                    False if self.config.subsample_mode.lower() != "conv" else True
                )
                shortcut = nn.Sequential(
                    DownSample(
                        down_scale=self.__down_scale,
                        in_channels=self.__in_channels,
                        out_channels=self.__in_channels,
                        norm=batch_norm,
                        mode=self.config.subsample_mode,
                        filt_size=self.config.get("filt_size", 3),  # for blur pool
                    ),
                    ZeroPadding(self.__in_channels, self.__out_channels),
                )
        else:
            shortcut = nn.Identity()
        return shortcut

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the block.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        out : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        identity = input

        out = self.main_stream(input)
        identity = self.shortcut(input)

        if identity.shape[-1] < out.shape[-1]:
            # downsampling using "avg" or "max" adopts no padding
            # hence shape[-1] of identity might be smaller by 1 than shape[-1] of out
            diff_sig_len = out.shape[-1] - identity.shape[-1]
            identity = F.pad(
                identity, [diff_sig_len // 2, diff_sig_len - diff_sig_len // 2]
            )

        out += identity
        out = self.out_activation(out)
        out = self.out_dropout(out)

        return out

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the block.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequenc
            Output shape of the block.

        """
        _seq_len = seq_len
        for module in self.main_stream:
            if isinstance(module, (nn.Dropout, nn.Dropout1d, nn.Identity)):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class ResNetBottleNeck(nn.Module, SizeMixin):
    """Bottleneck blocks for :class:`ResNet`.

    As for 1D ECG, should be of the "baby-giant-baby" pattern?

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input tensor.
    num_filters : Sequence[int]
        Number of filters for the neck conv layer.
    filter_length : int
        Lengths (sizes) of the filter kernels for the neck conv layer.
    subsample_length : int
        Subsample length, including pool size for short cut,
        and stride for the (top or neck) conv layer.
    groups : int, default 1
        Pattern of connections between inputs and outputs of the neck conv layer.
        For more details, ref. :class:`torch.nn.Conv1d`.
    dilation : int, default 1
        Dilation of the conv layers.
    base_width : numbers.Real, default 12*4
        Number of filters per group for the neck conv layer.
        Usually number of filters of the initial conv layer
        of the whole ResNet model.
    base_groups : int, default 1
        Pattern of connections between inputs and outputs of conv layers at the two ends,
        which should divide `groups`.
    base_filter_length : int, default 1
        Lengths (sizes) of the filter kernels for conv layers at the two ends.
    attn : dict, optional
        Attention mechanism for the neck conv layer.
        If is  None, no attention mechanism is used.
        Keys:

            - name: str,
              can be "se", "gc", "nl" (alias "nonlocal", "non-local"), etc.
            - pos: int,
              position of the attention mechanism.

        Other keys are specific to the attention mechanism.
    config : dict
        Other hyper-parameters, including
        increase channel method, subsample method, dropout,
        activation choices, weight initializer, and short cut patterns, etc.

    """

    __name__ = "ResNetBottleNeck"
    expansion = 4
    __DEFAULT_BASE_WIDTH__ = 12 * 4
    __DEFAULT_CONFIG__ = _DEFAULT_BLOCK_CONFIG.copy()
    __DEFAULT_CONFIG__["subsample_at"] = 0

    def __init__(
        self,
        in_channels: int,
        num_filters: int,
        filter_length: int,
        subsample_length: int,
        groups: int = 1,
        dilation: int = 1,
        base_width: Real = 12 * 4,
        base_groups: int = 1,
        base_filter_length: int = 1,
        attn: Optional[dict] = None,
        **config,
    ) -> None:
        super().__init__()
        self.__num_convs = 3
        self.config = CFG(self.__DEFAULT_CONFIG__.copy())
        self.config.update(config)
        self.expansion = self.config.get("expansion", self.expansion)

        self.__in_channels = in_channels
        # update denominator of computing neck_num_filters by init_num_filters
        self.__DEFAULT_BASE_WIDTH__ = self.config.get(
            "init_num_filters", self.__DEFAULT_BASE_WIDTH__
        )
        neck_num_filters = (
            int(num_filters * (base_width / self.__DEFAULT_BASE_WIDTH__)) * groups
        )
        self.__out_channels = [
            neck_num_filters,
            neck_num_filters,
            num_filters * self.expansion,
        ]
        self.__base_filter_length = base_filter_length
        self.__kernel_size = [
            base_filter_length,
            filter_length,
            base_filter_length,
        ]
        self.__down_scale = subsample_length
        self.__stride = subsample_length
        if groups % base_groups != 0:
            raise ValueError("`groups` should be divisible by `base_groups`")
        self.__base_groups = base_groups
        self.__groups = [
            base_groups,
            groups,
            base_groups,
        ]
        self.__attn = attn
        if self.__attn:
            self.__attn = CFG(self.__attn)

        if (
            self.config.increase_channels_method.lower() == "zero_padding"
            and self.__groups != 1
        ):
            raise ValueError(
                "zero padding for increasing channels can not be used with groups != 1"
            )

        self.__increase_channels = self.__out_channels[-1] > self.__in_channels
        self.shortcut = self._make_shortcut_layer()

        self.main_stream = nn.Sequential()
        conv_names = {
            0: "cba_head",
            1: "cba_neck",
            2: "cba_tail",
        }
        conv_in_channels = self.__in_channels
        dropout_config = self.config.get("dropout", 0)
        for i in range(self.__num_convs):
            if self.__attn and self.__attn["pos"] == i:
                self.main_stream.add_module(
                    self.__attn["name"],
                    make_attention_layer(conv_in_channels, **self.__attn),
                )
            conv_activation = (
                self.config.activation if i < self.__num_convs - 1 else None
            )
            conv_out_channels = self.__out_channels[i]
            conv_type = self.config.get("conv_type", None)
            if self.__kernel_size[i] == 1 and conv_type == "separable":
                conv_type = None
            self.main_stream.add_module(
                conv_names[i],
                Conv_Bn_Activation(
                    in_channels=conv_in_channels,
                    out_channels=conv_out_channels,
                    kernel_size=self.__kernel_size[i],
                    stride=(self.__stride if i == self.config.subsample_at else 1),
                    groups=self.__groups[i],
                    norm=True,
                    activation=conv_activation,
                    kw_activation=self.config.kw_activation,
                    kernel_initializer=self.config.kernel_initializer,
                    kw_initializer=self.config.kw_initializer,
                    bias=self.config.bias,
                    conv_type=conv_type,
                ),
            )
            if i < self.__num_convs - 1:
                if isinstance(dropout_config, dict):
                    if dropout_config["type"] == "1d" and dropout_config["p"] > 0:
                        self.main_stream.add_module(
                            f"dropout_{i}", nn.Dropout1d(dropout_config["p"])
                        )
                    elif dropout_config["type"] is None and dropout_config["p"] > 0:
                        self.main_stream.add_module(
                            f"dropout_{i}", nn.Dropout(dropout_config["p"])
                        )
                elif dropout_config > 0:  # float
                    self.main_stream.add_module(
                        f"dropout_{i}", nn.Dropout(dropout_config)
                    )
            conv_in_channels = conv_out_channels
        if self.__attn and self.__attn["pos"] == -1:
            self.main_stream.add_module(
                self.__attn["name"],
                make_attention_layer(conv_in_channels, **self.__attn),
            )

        if isinstance(self.config.activation, str):
            self.out_activation = Activations[self.config.activation.lower()](
                **self.config.kw_activation
            )
        else:
            self.out_activation = self.config.activation(**self.config.kw_activation)

        if isinstance(dropout_config, dict):
            if dropout_config["type"] == "1d" and dropout_config["p"] > 0:
                self.out_dropout = nn.Dropout1d(dropout_config["p"])
            elif dropout_config["type"] is None and dropout_config["p"] > 0:
                self.out_dropout = nn.Dropout(dropout_config["p"])
        elif dropout_config > 0:  # float
            self.out_dropout = nn.Dropout(self.config.dropout)
        else:
            self.out_dropout = nn.Identity()

    def _make_shortcut_layer(self) -> Union[nn.Module, None]:
        """Make shortcut layer for residual connection."""
        if self.__down_scale > 1 or self.__increase_channels:
            if self.config.increase_channels_method.lower() == "conv":
                shortcut = DownSample(
                    down_scale=self.__down_scale,
                    in_channels=self.__in_channels,
                    out_channels=self.__out_channels[-1],
                    groups=self.__base_groups,
                    norm=True,
                    mode=self.config.subsample_mode,
                    filt_size=self.config.get("filt_size", 3),  # for blur pool
                )
            elif self.config.increase_channels_method.lower() == "zero_padding":
                batch_norm = (
                    False if self.config.subsample_mode.lower() != "conv" else True
                )
                shortcut = nn.Sequential(
                    DownSample(
                        down_scale=self.__down_scale,
                        in_channels=self.__in_channels,
                        out_channels=self.__in_channels,
                        norm=batch_norm,
                        mode=self.config.subsample_mode,
                        filt_size=self.config.get("filt_size", 3),  # for blur pool
                    ),
                    ZeroPadding(self.__in_channels, self.__out_channels[-1]),
                )
        else:
            shortcut = nn.Identity()
        return shortcut

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the block.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        out : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        identity = input

        out = self.main_stream(input)
        identity = self.shortcut(input)

        if identity.shape[-1] < out.shape[-1]:
            # downsampling using "avg" or "max" adopts no padding
            # hence shape[-1] of identity might be smaller by 1 than shape[-1] of out
            diff_sig_len = out.shape[-1] - identity.shape[-1]
            identity = F.pad(
                identity, [diff_sig_len // 2, diff_sig_len - diff_sig_len // 2]
            )

        out += identity
        out = self.out_activation(out)
        out = self.out_dropout(out)

        return out

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the block.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequenc
            Output shape of the block.

        """
        _seq_len = seq_len
        for module in self.main_stream:
            if isinstance(module, (nn.Identity, nn.Dropout, nn.Dropout1d)):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class ResNetStem(nn.Sequential, SizeMixin):
    """The input stem of ResNet.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int or Sequence[int]
        Number of output channels.
    filter_lengths : int or Sequence[int]
        Length of the filter, or equivalently,
        kernel size(s) of the convolutions.
    conv_stride : int
        Stride of the convolution.
    pool_size : int
        Size of the pooling window.
    pool_stride : int
        Stride of the pooling window.
    subsample_mode : str
        Mode of subsampling, can be one of
        {:class:`DownSample`.__MODES__},
        or "s2d" (with aliases "space_to_depth", "SpaceToDepth").
    groups : int
        Number of groups for the convolution
    config : dict
        Other configs for convolution and pooling.

    """

    __name__ = "ResNetStem"

    def __init__(
        self,
        in_channels: int,
        out_channels: Union[int, Sequence[int]],
        filter_lengths: Union[int, Sequence[int]],
        conv_stride: int,
        pool_size: int,
        pool_stride: int,
        subsample_mode: str = "max",
        groups: int = 1,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__filter_lengths = filter_lengths
        if subsample_mode.lower() in ["s2d", "space_to_depth", "SpaceToDepth"]:
            self.add_module(
                "s2d",
                SpaceToDepth(
                    self.__in_channels, self.__out_channels, config.get("block_size", 4)
                ),
            )
            return
        if isinstance(self.__filter_lengths, int):
            self.__filter_lengths = [self.__filter_lengths]
        if isinstance(self.__out_channels, int):
            self.__out_channels = [self.__out_channels]
        assert len(self.__filter_lengths) == len(self.__out_channels)

        conv_in_channels = self.__in_channels
        for idx, fl in enumerate(self.__filter_lengths):
            self.add_module(
                f"conv_{idx}",
                Conv_Bn_Activation(
                    conv_in_channels,
                    self.__out_channels[idx],
                    self.__filter_lengths[idx],
                    stride=conv_stride if idx == 0 else 1,
                    groups=groups,
                    **config,
                ),
            )
            conv_in_channels = self.__out_channels[idx]
        if pool_stride > 1:
            self.add_module(
                "pool",
                DownSample(
                    pool_stride,
                    conv_in_channels,
                    kernel_size=pool_size,
                    groups=groups,
                    padding=(pool_stride - 1) // 2,
                    mode=subsample_mode.lower(),
                    **config,
                ),
            )

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the stem."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class ResNet(nn.Sequential, SizeMixin, CitationMixin):
    """ResNet model.

    The ResNet model is used for ECG classification by a team from
    the Stanford University [1]_, from which the application of
    deep learning to ECG analysis becomes more widely known and accepted.

    This implementation bases much on the torchvision implementation [2]_,
    although which is for image tasks.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input signal tensor.
    config : dict
        Hyper-parameters of the Module, ref. corr. config file.
        keyword arguments that must be set:

            - bias: bool,
              if True, each convolution will have a bias term.
            - num_blocks: sequence of int,
              number of building blocks in each macro block.
            - filter_lengths: int or sequence of int or sequence of sequences of int,
              filter length(s) (kernel size(s)) of the convolutions,
              with granularity to the whole network, to each macro block,
              or to each building block.
            - subsample_lengths: int or sequence of int or sequence of sequences of int,
              subsampling length(s) (ratio(s)) of all blocks,
              with granularity to the whole network, to each macro block,
              or to each building block,
              the former 2 subsample at the first building block.
            - groups: int,
              connection pattern (of channels) of the inputs and outputs.
            - stem: dict,
              other parameters that can be set for the input stem.
            - block: dict,
              other parameters that can be set for the building blocks.

        For a full list of configurable parameters, ref. corr. config file.

    References
    ----------
    .. [1] https://github.com/awni/ecg
    .. [2] https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py

    TODO
    ----
    1. Check performances of activations other than :class:`~torch.nn.ReLU`,
       especially :class:`~torch.nn.Mish`, etc.
    2. Add functionality of `replace_stride_with_dilation`.

    """

    __name__ = "ResNet"
    building_block = ResNetBasicBlock
    __DEFAULT_CONFIG__ = CFG(
        activation="relu",
        kw_activation={"inplace": True},
        kernel_initializer="he_normal",
        kw_initializer={},
        base_groups=1,
        dropouts=0,
    )

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        self.config.update(deepcopy(config))
        if not isinstance(self.config.get("building_block", None), str):
            # for those architectures that have multiple type of building blocks
            self.building_block = []
            self.additional_kw = []
            for b in self.config.building_block:
                if b.lower() in [
                    "bottleneck",
                    "bottle_neck",
                ]:
                    self.building_block.append(ResNetBottleNeck)
                    self.additional_kw.append(
                        CFG(
                            {
                                k: self.config[k]
                                for k in [
                                    "base_width",
                                    "base_groups",
                                    "base_filter_length",
                                ]
                                if k in self.config.keys()
                            }
                        )
                    )
                else:
                    self.building_block.append(ResNetBasicBlock)
                    self.additional_kw.append(CFG())
            assert isinstance(self.config.block, Sequence) and len(
                self.config.block
            ) == len(self.config.building_block) == len(self.config.num_blocks)
        elif self.config.get("building_block", "").lower() in [
            "bottleneck",
            "bottle_neck",
        ]:
            self.building_block = ResNetBottleNeck
            # additional parameters for bottleneck
            self.additional_kw = CFG(
                {
                    k: self.config[k]
                    for k in [
                        "base_width",
                        "base_groups",
                        "base_filter_length",
                    ]
                    if k in self.config.keys()
                }
            )
        else:
            self.additional_kw = CFG()

        stem_config = CFG(self.config.stem)
        stem_config.pop("num_filters", None)
        self.add_module(
            "input_stem",
            ResNetStem(
                in_channels=self.__in_channels,
                out_channels=self.config.stem.num_filters,
                # bottleneck use "base_groups"
                groups=self.config.get("base_groups", self.config.groups),
                activation=self.config.activation,
                **stem_config,
            ),
        )

        if isinstance(self.config.filter_lengths, int):
            self.__filter_lengths = list(
                repeat(self.config.filter_lengths, len(self.config.num_blocks))
            )
        else:
            self.__filter_lengths = self.config.filter_lengths
        assert len(self.__filter_lengths) == len(self.config.num_blocks), (
            f"`config.filter_lengths` indicates {len(self.__filter_lengths)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )
        if isinstance(self.config.subsample_lengths, int):
            self.__subsample_lengths = list(
                repeat(self.config.subsample_lengths, len(self.config.num_blocks))
            )
        else:
            self.__subsample_lengths = self.config.subsample_lengths
        assert len(self.__subsample_lengths) == len(self.config.num_blocks), (
            f"`config.subsample_lengths` indicates {len(self.__subsample_lengths)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )
        self.__num_filters = self.config.get(
            "num_filters",
            [
                (2**i) * self.input_stem.compute_output_shape()[1]
                for i in range(len(self.config.num_blocks))
            ],
        )
        assert len(self.__num_filters) == len(self.config.num_blocks), (
            f"`config.num_filters` indicates {len(self.__num_filters)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )
        if isinstance(self.config.dropouts, (Real, dict)):
            self.__dropouts = list(
                repeat(self.config.dropouts, len(self.config.num_blocks))
            )
        else:
            self.__dropouts = self.config.dropouts
        assert len(self.__dropouts) == len(self.config.num_blocks), (
            f"`config.dropouts` indicates {len(self.__dropouts)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )

        # grouped resnet (basic) blocks,
        # number of channels are doubled at the first block of each macro-block
        macro_in_channels = self.input_stem.compute_output_shape()[1]
        for macro_idx, nb in enumerate(self.config.num_blocks):
            macro_num_filters = self.__num_filters[macro_idx]
            macro_filter_lengths = self.__filter_lengths[macro_idx]
            macro_subsample_lengths = self.__subsample_lengths[macro_idx]
            block_in_channels = macro_in_channels
            block_num_filters = macro_num_filters
            if isinstance(macro_filter_lengths, int):
                block_filter_lengths = list(repeat(macro_filter_lengths, nb))
            else:
                block_filter_lengths = macro_filter_lengths
            assert len(block_filter_lengths) == nb, (
                f"at the {macro_idx}-th macro block, `macro_subsample_lengths` indicates {len(macro_subsample_lengths)} "
                f"building blocks, while `config.num_blocks[{macro_idx}]` indicates {nb}"
            )
            if isinstance(macro_subsample_lengths, int):
                # subsample at the first building block
                block_subsample_lengths = list(repeat(1, nb))
                block_subsample_lengths[0] = macro_subsample_lengths
            else:
                block_subsample_lengths = macro_subsample_lengths
            assert len(block_subsample_lengths) == nb, (
                f"at the {macro_idx}-th macro block, `macro_subsample_lengths` indicates {len(macro_subsample_lengths)} "
                f"building blocks, while `config.num_blocks[{macro_idx}]` indicates {nb}"
            )
            if isinstance(self.building_block, Sequence):
                bb = self.building_block[macro_idx]
                bb_config = self.config.block[macro_idx]
                bb_kw = self.additional_kw[macro_idx]
            else:
                bb = self.building_block
                bb_config = self.config.block
                bb_kw = self.additional_kw
            for block_idx in range(nb):
                dropout = bb_config.pop("dropout", self.__dropouts[macro_idx])
                self.add_module(
                    f"{bb.__name__}_{macro_idx}_{block_idx}",
                    bb(
                        in_channels=block_in_channels,
                        num_filters=block_num_filters,
                        filter_length=block_filter_lengths[block_idx],
                        subsample_length=block_subsample_lengths[block_idx],
                        groups=self.config.groups,
                        dilation=1,
                        dropout=dropout,
                        **(bb_kw),
                        **(bb_config),
                    ),
                )
                block_in_channels = block_num_filters * bb.expansion
            macro_in_channels = macro_num_filters * bb.expansion

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(
            set(
                self.config.get("doi", [])
                + [
                    "10.1109/cvpr.2016.90",
                    "10.1038/s41591-018-0268-3",
                    "10.1038/s41467-020-15432-4",
                    "10.1088/1361-6579/ac6aa3",
                ]
            )
        )



"""
"""

import warnings
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional, Sequence, Union, Dict

import numpy as np
from torch import Tensor

from ..utils.misc import ReprMixin, add_docstring
from ..utils.utils_data import ECGWaveFormNames
from ..utils.utils_metrics import (
    QRS_score,
    cls_to_bin,
    metrics_from_confusion_matrix,
    confusion_matrix,
    ovr_confusion_matrix,
    compute_wave_delineation_metrics,
)


__all__ = [
    "Metrics",
    "ClassificationMetrics",
    "RPeaksDetectionMetrics",
    "WaveDelineationMetrics",
]


class Metrics(ReprMixin, ABC):
    """Base class for metrics."""

    __name__ = "Metrics"

    def __call__(self, *args: Any, **kwargs: Any) -> "Metrics":
        return self.compute(*args, **kwargs)

    @abstractmethod
    def compute(self, *args: Any, **kwargs: Any) -> "Metrics":
        raise NotImplementedError


class ClassificationMetrics(Metrics):
    """Metrics for the task of classification.

    Parameters
    ----------
    multi_label : bool, default True
        Whether is multi-label classification task.
    macro : bool, default True
        Whether to use macro-averaged metrics.
    extra_metrics : callable, optional
        Extra metrics to compute,
        has to be a function with signature:

        .. code-block:: python

            def extra_metrics(
                labels : np.ndarray
                outputs : np.ndarray
                num_classes : Optional[int]=None
                weights : Optional[np.ndarray]=None
            ) -> dict

    """

    __name__ = "ClassificationMetrics"

    def __init__(
        self,
        multi_label: bool = True,
        macro: bool = True,
        extra_metrics: Optional[Callable] = None,
    ) -> None:
        self.multi_label = multi_label
        self.set_macro(macro)
        self._extra_metrics = extra_metrics
        self._em = {}
        self._metrics = {
            k: np.nan
            for k in [
                "sens",  # sensitivity, recall, hit rate, or true positive rate
                "spec",  # specificity, selectivity, or true negative rate
                "prec",  # precision, or positive predictive value
                "npv",  # negative predictive value
                "jac",  # jaccard index, threat score, or critical success index
                "acc",  # accuracy
                "phi",  # phi coefficient, or matthews correlation coefficient
                "fnr",  # false negative rate, or miss rate
                "fpr",  # false positive rate, or fall-out
                "fdr",  # false discovery rate
                "for",  # false omission rate
                "plr",  # positive likelihood ratio
                "nlr",  # negative likelihood ratio
                "pt",  # prevalence threshold
                "ba",  # balanced accuracy
                "f1",  # f1-measure
                "fm",  # fowlkes-mallows index
                "bm",  # bookmaker informedness
                "mk",  # markedness
                "dor",  # diagnostic odds ratio
                "auroc",  # area under the receiver-operater characteristic curve (ROC AUC)
                "auprc",  # area under the precision-recall curve
            ]
        }
        self._metrics.update({f"macro_{k}": np.nan for k in self._metrics})
        self._cm = None
        self._cm_ovr = None

    def set_macro(self, macro: bool) -> None:
        """Set whether to use macro-averaged metrics.

        Parameters
        ----------
        macro : bool
            Whether to use macro-averaged metrics.

        """
        self.__prefix = ""
        self.macro = macro
        if macro:
            self.__prefix = "macro_"

    @add_docstring(
        metrics_from_confusion_matrix.__doc__.replace(
            "metrics : dict", f"self : {__name__},"
        )
        .replace(
            "Metrics computed from the one-vs-rest confusion matrix.",
            "The metrics object itself with the computed metrics.",
        )
        .replace(
            "metrics = metrics_from_confusion_matrix(labels, outputs)",
            """metrics = ClassificationMetrics()
    >>> metrics = metrics.compute(labels, outputs)
    >>> metrics.fl_measure
    0.5062821146226457
    >>> metrics.set_macro(False)
    >>> metrics.fl_measure
    array([0.46938776, 0.4742268 , 0.4375    , 0.52941176, 0.58      ,
       0.57692308, 0.55769231, 0.48351648, 0.55855856, 0.3956044 ])""",
        )
    )
    def compute(
        self,
        labels: Union[np.ndarray, Tensor],
        outputs: Union[np.ndarray, Tensor],
        num_classes: Optional[int] = None,
        weights: Optional[np.ndarray] = None,
        thr: float = 0.5,
    ) -> "ClassificationMetrics":
        labels, outputs = cls_to_bin(labels, outputs, num_classes)
        num_samples, num_classes = np.shape(labels)
        # probability outputs to binary outputs
        bin_outputs = np.zeros_like(outputs, dtype=int)
        bin_outputs[outputs >= thr] = 1
        bin_outputs[outputs < thr] = 0
        self._cm = confusion_matrix(labels, bin_outputs, num_classes)
        self._cm_ovr = ovr_confusion_matrix(labels, bin_outputs, num_classes)
        self._metrics = metrics_from_confusion_matrix(
            labels, outputs, num_classes, weights
        )
        if self._extra_metrics is not None:
            self._em = self._extra_metrics(labels, outputs, num_classes, weights)
            self._metrics.update(self._em)

        return self

    @add_docstring(compute.__doc__)
    def __call__(
        self,
        labels: Union[np.ndarray, Tensor],
        outputs: Union[np.ndarray, Tensor],
        num_classes: Optional[int] = None,
        weights: Optional[np.ndarray] = None,
        thr: float = 0.5,
    ) -> "ClassificationMetrics":
        return self.compute(labels, outputs, num_classes, weights)

    @property
    def sensitivity(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}sens"]

    @property
    def recall(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}sens"]

    @property
    def hit_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}sens"]

    @property
    def true_positive_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}sens"]

    @property
    def specificity(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}spec"]

    @property
    def selectivity(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}spec"]

    @property
    def true_negative_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}spec"]

    @property
    def precision(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}prec"]

    @property
    def positive_predictive_value(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}prec"]

    @property
    def negative_predictive_value(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}npv"]

    @property
    def jaccard_index(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}jac"]

    @property
    def threat_score(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}jac"]

    @property
    def critical_success_index(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}jac"]

    @property
    def accuracy(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}acc"]

    @property
    def phi_coefficient(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}phi"]

    @property
    def matthews_correlation_coefficient(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}phi"]

    @property
    def false_negative_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}fnr"]

    @property
    def miss_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}fnr"]

    @property
    def false_positive_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}fpr"]

    @property
    def fall_out(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}fpr"]

    @property
    def false_discovery_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}fdr"]

    @property
    def false_omission_rate(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}for"]

    @property
    def positive_likelihood_ratio(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}plr"]

    @property
    def negative_likelihood_ratio(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}nlr"]

    @property
    def prevalence_threshold(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}pt"]

    @property
    def balanced_accuracy(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}ba"]

    @property
    def f1_measure(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}f1"]

    @property
    def fowlkes_mallows_index(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}fm"]

    @property
    def bookmaker_informedness(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}bm"]

    @property
    def markedness(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}mk"]

    @property
    def diagnostic_odds_ratio(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}dor"]

    @property
    def area_under_the_receiver_operater_characteristic_curve(
        self,
    ) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}auroc"]

    @property
    def auroc(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}auroc"]

    @property
    def area_under_the_precision_recall_curve(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}auprc"]

    @property
    def auprc(self) -> Union[float, np.ndarray]:
        return self._metrics[f"{self.__prefix}auprc"]

    @property
    def classification_report(self) -> dict:
        if self.__prefix == "macro_":
            return {
                k.replace("macro_", ""): v
                for k, v in self._metrics.items()
                if k.startswith("macro_")
            }
        else:
            return {
                k: v for k, v in self._metrics.items() if not k.startswith("macro_")
            }

    @property
    def extra_metrics(self) -> dict:
        return self._em

    def extra_repr_keys(self) -> List[str]:
        return [
            "multi_label",
            "macro",
        ]


class RPeaksDetectionMetrics(Metrics):
    """Metrics for the task of R peaks detection,
    as proposed in CPSC2019.

    Parameters
    ----------
    thr : float, default 0.075
        Threshold for a prediction to be truth positive,
        with units in seconds,
    extra_metrics : callable, optional
        Extra metrics to compute,
        has to be a function with signature

        .. code-block:: python

            def extra_metrics(
                labels : Sequence[Union[Sequence[int], np.ndarray]],
                outputs : Sequence[Union[Sequence[int], np.ndarray]],
                fs : int
            ) -> dict

    """

    __name__ = "RPeaksDetectionMetrics"

    def __init__(
        self,
        thr: float = 0.075,
        extra_metrics: Optional[Callable] = None,
    ) -> None:
        self.thr = thr
        self._extra_metrics = extra_metrics
        self._em = {}
        self._metrics = {"qrs_score": np.nan}

    @add_docstring(
        QRS_score.__doc__.replace("rpeaks_truths", "labels")
        .replace("rpeaks_preds", "outputs")
        .replace(
            "thr : float, default 0.075",
            "thr : float, optional, defaults to `self.thr`",
        )
        .replace("rec_acc : float", f"self : {__name__}")
        .replace(
            "Accuracy of predictions.",
            "The metrics object itself with the computed metrics.",
        )
        .rstrip(" \n")
        + """

        Examples
        --------
        >>> labels = [np.array([500, 1000])]
        >>> outputs = [np.array([500, 700, 1000])]  # a false positive at 700
        >>> metrics = RPeaksDetectionMetrics()
        >>> metrics = metrics.compute(labels, outputs, fs=500)
        >>> metrics.qrs_score
        0.7

        """
    )
    def compute(
        self,
        labels: Sequence[Union[Sequence[int], np.ndarray]],
        outputs: Sequence[Union[Sequence[int], np.ndarray]],
        fs: int,
        thr: Optional[float] = None,
    ) -> "RPeaksDetectionMetrics":
        self._metrics["qrs_score"] = QRS_score(labels, outputs, fs, thr or self.thr)
        if self._extra_metrics is not None:
            self._em = self._extra_metrics(labels, outputs, fs)
            self._metrics.update(self._em)

        return self

    @add_docstring(compute.__doc__)
    def __call__(
        self,
        labels: Sequence[Union[Sequence[int], np.ndarray]],
        outputs: Sequence[Union[Sequence[int], np.ndarray]],
        fs: int,
        thr: Optional[float] = None,
    ) -> "RPeaksDetectionMetrics":
        return self.compute(labels, outputs, fs, thr)

    @property
    def qrs_score(self) -> float:
        return self._metrics["qrs_score"]

    @property
    def extra_metrics(self) -> dict:
        return self._em

    def extra_repr_keys(self) -> List[str]:
        return ["thr"]


class WaveDelineationMetrics(Metrics):
    """Metrics for the task of ECG wave delineation.

    Parameters
    ----------
    macro : bool, default True
        Whether to use macro-averaged metrics or not.
    tol : float, default 0.15
        Tolerance for the duration of the waveform,
        with units in seconds.
    extra_metrics : callable, optional
        Extra metrics to compute,
        has to be a function with signature

        .. code-block:: python

            def extra_metrics(
                labels: Sequence[Union[Sequence[int], np.ndarray]],
                outputs: Sequence[Union[Sequence[int], np.ndarray]],
                fs: int
            ) -> dict

    """

    __name__ = "WaveDelineationMetrics"

    def __init__(
        self,
        macro: bool = True,
        tol: float = 0.15,
        extra_metrics: Optional[Callable] = None,
    ) -> None:
        self.set_macro(macro)
        self.tol = tol
        self._extra_metrics = extra_metrics
        self._em = {}
        self._metrics = {
            k: None
            for k in [
                "sensitivity",
                "precision",
                "f1_score",
                "mean_error",
                "standard_deviation",
                "jaccard",
            ]
        }
        self._metrics.update({f"macro_{k}": np.nan for k in self._metrics})

    def set_macro(self, macro: bool) -> None:
        """Set whether to use macro-averaged metrics or not.

        Parameters
        ----------
        macro : bool
            Shether to use macro-averaged metrics.

        """
        self.__prefix = ""
        self.macro = macro
        if macro:
            self.__prefix = "macro_"

    @add_docstring(
        f"""
        Compute metrics for the task of ECG wave delineation
        (sensitivity, precision, f1_score, mean error and standard deviation of the mean errors)
        for multiple evaluations.

        Parameters
        ----------
        labels : numpy.ndarray or torch.Tensor
            Ground truth masks,
            of shape ``(n_samples, n_channels, n_timesteps)``.
        outputs : numpy.ndarray or torch.Tensor
            Predictions corresponding to `labels`,
            of the same shape.
        class_map : dict
            Class map, mapping names to waves to numbers from 0 to n_classes-1,
            the keys should contain {", ".join([f'"{item}"' for item in ECGWaveFormNames])}.
        fs : numbers.Real
            Sampling frequency of the signal corresponding to the masks,
            used to compute the duration of each waveform,
            and thus the error and standard deviations of errors.
        mask_format : str, default "channel_first"
            Format of the mask, one of the following:
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first").
        tol : float, optional
            Tolerance for the duration of the waveform,
            with units in seconds.
            Defaults to `self.tol`.

        Returns
        -------
        self : WaveDelineationMetrics
            The metrics object itself with the computed metrics.

        """
    )
    def compute(
        self,
        labels: Union[np.ndarray, Tensor],
        outputs: Union[np.ndarray, Tensor],
        class_map: Dict[str, int],
        fs: int,
        mask_format: str = "channel_first",
        tol: Optional[float] = None,
    ) -> "WaveDelineationMetrics":
        # wave delineation specific metrics
        truth_masks = labels.numpy() if isinstance(labels, Tensor) else labels
        pred_masks = outputs.numpy() if isinstance(outputs, Tensor) else outputs
        raw_metrics = compute_wave_delineation_metrics(
            truth_masks, pred_masks, class_map, fs, mask_format, tol or self.tol
        )
        self._metrics = {
            metric: {
                f"{wf}_{pos}": raw_metrics[f"{wf}_{pos}"][metric]
                for wf in class_map
                for pos in [
                    "onset",
                    "offset",
                ]
            }
            for metric in [
                "sensitivity",
                "precision",
                "f1_score",
                "mean_error",
                "standard_deviation",
            ]
        }
        self._metrics.update(
            {
                f"macro_{metric}": np.nanmean(list(self._metrics[metric].values()))
                for metric in [
                    "sensitivity",
                    "precision",
                    "f1_score",
                    "mean_error",
                    "standard_deviation",
                ]
            }
        )
        # sample-wise metrics
        warnings.simplefilter("ignore")
        clf_mtx = ClassificationMetrics()
        swm = clf_mtx.compute(
            truth_masks.reshape((-1)).copy(),
            pred_masks.reshape((-1)).copy(),
            max(class_map.values()) + 1,
        )
        warnings.simplefilter("default")
        self._metrics.update(
            {
                "jaccard": {k: swm._metrics["jac"][v] for k, v in class_map.items()},
                "macro_jaccard": swm._metrics["macro_jac"],
            }
        )

        return self

    @add_docstring(compute.__doc__)
    def __call__(
        self,
        labels: Union[np.ndarray, Tensor],
        outputs: Union[np.ndarray, Tensor],
        class_map: Dict[str, int],
        fs: int,
        mask_format: str = "channel_first",
        tol: Optional[float] = None,
    ) -> "WaveDelineationMetrics":
        return self.compute(labels, outputs, class_map, fs, mask_format, tol)

    @property
    def sensitivity(self) -> Union[float, Dict[str, float]]:
        return self._metrics[f"{self.__prefix}sensitivity"]

    @property
    def precision(self) -> Union[float, Dict[str, float]]:
        return self._metrics[f"{self.__prefix}precision"]

    @property
    def f1_score(self) -> Union[float, Dict[str, float]]:
        return self._metrics[f"{self.__prefix}f1_score"]

    @property
    def mean_error(self) -> Union[float, Dict[str, float]]:
        return self._metrics[f"{self.__prefix}mean_error"]

    @property
    def standard_deviation(self) -> Union[float, Dict[str, float]]:
        return self._metrics[f"{self.__prefix}standard_deviation"]

    @property
    def jaccard_index(self) -> Union[float, Dict[str, float]]:
        return self._metrics[f"{self.__prefix}jaccard"]

    @property
    def extra_metrics(self) -> dict:
        return self._em

    def extra_repr_keys(self) -> List[str]:
        return ["tol"]
